/*
 * arch/arm/mach-kirkwood/wtplug-setup.c
 *
 * Promwad WTPlug Reference Board Setup
 *
 * This file is licensed under the terms of the GNU General Public
 * License version 2.  This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/mtd/partitions.h>
#include <linux/mv643xx_eth.h>
#include <linux/i2c.h>
#include <linux/gpio.h>
#include <linux/gpio_keys.h>
#include <linux/input.h>
#include <sound/alc5623.h>
#include <linux/leds.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <mach/kirkwood.h>
#include <linux/delay.h>
#include "common.h"
#include "mpp.h"

/* for the PCIe reset workaround */
#include <plat/pcie.h>

#ifdef CONFIG_WTPLUG_THINCLIENT
#define WTPLUG_BOARD_NAME		"Promwad Thin Client"
#define WTPLUG_GPIO_RED_LED		17
#define WTPLUG_GPIO_GREEN_LED		7
#define WTPLUG_GPIO_PEX0_RST		13
#define WTPLUG_GPIO_GPU_RST		38
#else
#define WTPLUG_BOARD_NAME		"Promwad Plug Computer"
#define WTPLUG_GPIO_RED_LED		39
#define WTPLUG_GPIO_GREEN_LED		38
#define WTPLUG_GPIO_PEX0_RST		42
#endif

#ifdef CONFIG_WTPLUG_NAND_HYNIX	
#define WTPLUG_NAND_CHIPDELAY		400
#define WTPLUG_NAND_UBOOT_SZ		SZ_4M
#else
#define WTPLUG_NAND_CHIPDELAY		25
#define WTPLUG_NAND_UBOOT_SZ		SZ_1M
#endif


#ifdef CONFIG_WTPLUG_THINCLIENT
static struct mtd_partition wtplug_nand_parts[] = {
	{
		.name = "u-boot",
		.offset = 0,
		.size = WTPLUG_NAND_UBOOT_SZ
	}, {
		.name = "uImage",
		.offset = MTDPART_OFS_NXTBLK,
		.size = SZ_4M
	}, {
		.name = "uImage_rcvr",
		.offset = MTDPART_OFS_NXTBLK,
		.size = SZ_4M
	}, {
		.name = "rootfs_rcvr",
		.offset = MTDPART_OFS_NXTBLK,
		.size = SZ_64M		/* TODO: clarify */
	}, {
		.name = "rootfs",
		.offset = MTDPART_OFS_NXTBLK,
		.size = MTDPART_SIZ_FULL
	},
};
#else
static struct mtd_partition wtplug_nand_parts[] = {
	{
		.name = "u-boot",
		.offset = 0,
		.size = WTPLUG_NAND_UBOOT_SZ
	}, {
		.name = "uImage",
		.offset = MTDPART_OFS_NXTBLK,
		.size = SZ_4M
	}, {
		.name = "auxfs",
		.offset = MTDPART_OFS_NXTBLK,
		.size = SZ_2M * 100
	}, {
		.name = "rootfs",
		.offset = MTDPART_OFS_NXTBLK,
		.size = MTDPART_SIZ_FULL		
	},
};
#endif

static struct mv643xx_eth_platform_data wtplug_ge00_data = {
	.phy_addr	= MV643XX_ETH_PHY_ADDR(1),
};

static struct mv643xx_eth_platform_data wtplug_ge01_data = {
	.phy_addr	= MV643XX_ETH_PHY_ADDR(0),
};

#ifndef CONFIG_WTPLUG_THINCLIENT
static struct gpio_led wtplug_led_pins[] = {
	{
		.name			= "wtplug:red:wmode",
		.default_trigger	= "phy0tx",
		.gpio			= WTPLUG_GPIO_RED_LED,
		.active_low		= 0,
	},
	{
		.name			= "wtplug:green:health",
		.default_trigger	= "heartbeat",
		.gpio			= WTPLUG_GPIO_GREEN_LED,
		.active_low		= 0,
	},
};

static struct gpio_led_platform_data wtplug_led_data = {
	.leds		= wtplug_led_pins,
	.num_leds	= ARRAY_SIZE(wtplug_led_pins),
};

static struct platform_device wtplug_leds = {
	.name	= "leds-gpio",
	.id	= -1,
	.dev	= {
		.platform_data	= &wtplug_led_data,
	}
};
#endif

#ifdef CONFIG_WTPLUG_THINCLIENT
static unsigned int wtplug_mpp_config[] __initdata = {
	MPP7_GPO,	/* LED_S_G */
	MPP8_TW0_SDA,
	MPP9_TW0_SCK,
	MPP13_GPIO,	/* PEX0_RST */		
	MPP14_GPIO,	/* WLAN_DIS */
	MPP15_GPIO,	/* UA0_RTS */
	MPP16_GPIO,	/* UA0_CTS */	
	MPP17_GPIO,	/* LED_S_R */
	MPP28_GPIO,	/* BUTTON */
	MPP29_GPIO,	/* LED_WLAN */
	MPP35_GPIO,	/* PEX0_CLKREQ */
	MPP38_GPIO,	/* GPU_RESET */	
	MPP39_AU_I2SBCLK,
	MPP40_AU_I2SDO,
	MPP43_AU_I2SDI,
	MPP41_AU_I2SLRCLK,
	MPP42_AU_I2SMCLK,	
	MPP44_GPIO,	/* WAKE_WIFI */	
	MPP45_GPIO,	/* USB_IND */	
	MPP46_GPIO,	/* USB_HS_IND */
	MPP47_GPIO,	/* RST_GE */
	MPP48_GPIO,	/* RST_USB */
	0
};
#else
static unsigned int wtplug_mpp_config[] __initdata = {
	MPP8_GPIO,	/* UA0_RTS */
	MPP9_GPIO,	/* UA0_CTS */	
	MPP35_GPIO,	/* PEX0_CLKREQ */
	MPP38_GPIO,	/* LED_S_R */
	MPP39_GPIO,	/* LED_S_G */
	MPP40_GPIO,	/* LED_PLAN */	
	MPP41_GPIO,	/* LED_WLAN */	
	MPP42_GPIO,	/* PEX0_RST */	
	MPP43_GPIO,	/* WLAN_DIS */	
	MPP44_GPIO,	/* WAKE_WIFI */	
	MPP45_GPIO,	/* USB_IND */	
	MPP46_GPIO,	/* USB_HS_IND */
	MPP47_GPIO,	/* RST_GE */
	MPP48_GPIO,	/* RST_USB */
	0
};
#endif

#ifdef CONFIG_WTPLUG_THINCLIENT
static struct alc5623_platform_data alc5621_data = {
	.add_ctrl = 0x3700,
	.jack_det_ctrl = 0x4980,
};

static struct i2c_board_info i2c0_board_info[] __initdata = {
	{
		I2C_BOARD_INFO("alc5621", 0x1a),
		.platform_data = &alc5621_data,
	},
};
#endif

/*
static struct i2c_board_info i2c1_board_info[] __initdata = {
	{
		I2C_BOARD_INFO("usb2514", 0x58),
	},
};
*/

static void __init wtplug_init(void)
{
	/*
	 * Basic setup. Needs to be called early.
	 */
	kirkwood_init();
	kirkwood_mpp_conf(wtplug_mpp_config);

	kirkwood_uart0_init();
	kirkwood_nand_init(ARRAY_AND_SIZE(wtplug_nand_parts),
			   WTPLUG_NAND_CHIPDELAY);

	kirkwood_i2c_init();
/*
	i2c_register_board_info(1, i2c1_board_info,
			ARRAY_SIZE(i2c1_board_info));
*/
	
/*
	if (gpio_request(48, "USB Hub Enable") != 0 ||
	    gpio_direction_output(48, 1) != 0)
		printk(KERN_ERR "can't set up GPIO 48 (USB Hub Enable)\n");	
*/
	kirkwood_ehci_init();
	kirkwood_ge00_init(&wtplug_ge00_data);
	kirkwood_ge01_init(&wtplug_ge01_data);

#ifndef CONFIG_WTPLUG_THINCLIENT
	platform_device_register(&wtplug_leds);
#endif
#ifdef CONFIG_WTPLUG_THINCLIENT
	i2c_register_board_info(0, i2c0_board_info,
		ARRAY_SIZE(i2c0_board_info));
	kirkwood_audio_init();
#endif	
}

#ifdef CONFIG_WTPLUG_THINCLIENT
static int wtplug_thinclient_reset_gpu(void)
{
	if (gpio_request(WTPLUG_GPIO_GPU_RST, "GPU reset")) {
		printk(KERN_ERR "unable to allocate GPU reset GPIO %d\n",
				WTPLUG_GPIO_GPU_RST);
		return 1;
	}

	if (gpio_direction_output(WTPLUG_GPIO_GPU_RST, 0)) {
		printk(KERN_ERR "unable to set direction of GPU reset GPIO %d",
				WTPLUG_GPIO_GPU_RST);
		return 1;
	}

	mdelay(200);

	gpio_set_value(WTPLUG_GPIO_GPU_RST, 1);

	return 0;
}
#endif

static int __init wtplug_pci_init(void)
{
	if (machine_is_wtplug()) {
		if (gpio_request(WTPLUG_GPIO_PEX0_RST, "PCI Express Enable") != 0 ||
		    gpio_direction_output(WTPLUG_GPIO_PEX0_RST, 1) != 0)
			printk(KERN_ERR "can't set up GPIO %d (PCI Express Enable)\n",
			  WTPLUG_GPIO_PEX0_RST);
#ifdef CONFIG_WTPLUG_THINCLIENT
		wtplug_thinclient_reset_gpu();
#endif
		orion_pcie_reset((void __iomem *)PCIE_VIRT_BASE);
#ifdef CONFIG_WTPLUG_THINCLIENT
		orion_pcie_reset((void __iomem *)PCIE1_VIRT_BASE);
		kirkwood_pcie_init(KW_PCIE0 | KW_PCIE1);
		kirkwood_pcie_tune(KW_PCIE0 | KW_PCIE1);
#else
		kirkwood_pcie_init(KW_PCIE0);
#endif
	}
	return 0;
}
subsys_initcall(wtplug_pci_init);

MACHINE_START(WTPLUG, WTPLUG_BOARD_NAME)
	/* Maintainer: Viktar Palstsiuk <viktar.palstsiuk@promwad.com> */
	.boot_params	= 0x00000100,
	.init_machine	= wtplug_init,
	.map_io		= kirkwood_map_io,
	.init_early     = kirkwood_init_early,
	.init_irq	= kirkwood_init_irq,
	.timer		= &kirkwood_timer,
MACHINE_END
