/*
 * (C) Copyright 2009
 * Marvell Semiconductor <www.marvell.com>
 * Written-by: Siddarth Gore <gores@marvell.com>
 *
 * Modified-by: Viktar Palstsiuk <viktar.palstsiuk@promwad.com>
 * 
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 */

#ifndef _CONFIG_WTPLUG_H
#define _CONFIG_WTPLUG_H

/*
 * Version number information
 */
#ifdef CONFIG_BOARD_IS_PROMWAD_THINCLIENT
# define CONFIG_IDENT_STRING	"\nPromwad-ThinClient"
#else
# define CONFIG_IDENT_STRING	"\nPromwad-PlugComputer"
#endif

/*
 * High Level Configuration Options (easy to change)
 */
#define CONFIG_SHEEVA_88SV131	1	/* CPU Core subversion */
#define CONFIG_KIRKWOOD		1	/* SOC Family Name */
#define CONFIG_KW88F6281	1	/* SOC Name */
#define CONFIG_MACH_WTPLUG		/* Machine type */
#define CONFIG_SKIP_LOWLEVEL_INIT	/* disable board lowlevel_init */

#if !defined(CONFIG_BOARD_IS_PROMWAD_THINCLIENT)
# define CONFIG_NAND_BB_AUTO	1	/* automatic create nand badblock */
#endif

/*
 * Commands configuration
 */
#define CONFIG_SYS_NO_FLASH		/* Declare no flash (NOR/SPI) */
#include <config_cmd_default.h>
#define CONFIG_CMD_DHCP
#define CONFIG_CMD_ENV
#define CONFIG_CMD_FAT
#define CONFIG_CMD_EXT2		/* support ext2 file system */
#define CONFIG_CMD_NAND
#define CONFIG_CMD_PING
#define CONFIG_CMD_USB
#define CONFIG_EHCI_DCACHE 1
#define CONFIG_CMD_DATE
#ifdef CONFIG_BOARD_IS_PROMWAD_THINCLIENT
#define CONFIG_CMD_I2C
#endif

/**
 * shell options
 */
#define CONFIG_SYS_HUSH_PARSER 1
#define CONFIG_SYS_PROMPT_HUSH_PS2	"> "

/*
 * mv-common.h should be defined after CMD configs since it used them
 * to enable certain macros
 */
#include "mv-common.h"

/*
 *  Environment variables configurations
 */
#ifdef CONFIG_CMD_NAND
#define CONFIG_CMD_NAND_TRIMFFS		1
#define CONFIG_ENV_IS_IN_NAND		1
#ifdef CONFIG_NAND_HYNIX
#define CONFIG_ENV_SECT_SIZE		0x200000	/* 2M */
#else
#define CONFIG_ENV_SECT_SIZE		0x40000	/* 256K */
#endif
#else
#define CONFIG_ENV_IS_NOWHERE		1	/* if env in SDRAM */
#endif
/*
 * max 4k env size is enough, but in case of nand
 * it has to be rounded to sector size
 */
#ifdef CONFIG_NAND_HYNIX
#define CONFIG_ENV_SIZE			0x200000	/* 2M */
#define CONFIG_ENV_ADDR			0x200000
#define CONFIG_ENV_OFFSET		0x200000	/* env starts here */
#else
#define CONFIG_ENV_SIZE			0x40000	/* 256k */
#define CONFIG_ENV_ADDR			0xC0000
#define CONFIG_ENV_OFFSET		0xC0000	/* env starts here */
#endif

/*
 * Default environment variables
 */
#ifdef CONFIG_BOARD_IS_PROMWAD_THINCLIENT
 #define CONFIG_BOOTCOMMAND \
	"tc_rootfs_update; " \
	"setenv bootargs ${x_bootargs} ${x_bootargs_root}; " \
	"run x_bootcmd_kernel; " \
	"bootm 0x6400000; "

 #ifdef CONFIG_NAND_HYNIX
  #define CONFIG_EXTRA_ENV_SETTINGS \
	"ipaddr=192.168.10.234\0" \
	"serverip=192.168.10.150\0" \
	"x_bootcmd_kernel=nand read.e 0x6400000 0x400000 0x400000\0" \
	"x_bootcmd_kernel_rcvr=nand read.e 0x6400000 0x800000 0x400000\0" \
	"x_bootargs=console=ttyS0,115200\0" \
	"x_bootargs_root=ubi.mtd=4 root=ubi0:rootfs rootfstype=ubifs\0" \
	"x_bootargs_root_rcvr=ubi.mtd=3 root=ubi0:rootfs rootfstype=ubifs\0" \
	"bootcmd_rcvr=setenv bootargs ${x_bootargs_root_rcvr} quiet; run x_bootcmd_kernel_rcvr; bootm 0x6400000; \0"
 #else
  #define CONFIG_EXTRA_ENV_SETTINGS \
	"ipaddr=192.168.10.234\0" \
	"serverip=192.168.10.150\0" \
	"x_bootcmd_kernel=nand read.e 0x6400000 0x100000 0x400000\0" \
	"x_bootcmd_kernel_rcvr=nand read.e 0x6400000 0x500000 0x400000\0" \
	"x_bootargs=console=ttyS0,115200\0" \
	"x_bootargs_root=ubi.mtd=4 root=ubi0:rootfs rootfstype=ubifs\0" \
	"x_bootargs_root_rcvr=ubi.mtd=3 root=ubi0:rootfs rootfstype=ubifs\0" \
	"bootcmd_rcvr=setenv bootargs ${x_bootargs_root_rcvr} quiet; run x_bootcmd_kernel_rcvr; bootm 0x6400000; \0"
 #endif
#else
 #define CONFIG_BOOTCOMMAND \
	"usb start; "\
	"if run x_usbcmd_kernel; then " \
		"setenv bootargs ${x_bootargs} ${x_usbarg_root}; " \
	"else " \
		"run x_bootcmd_kernel; " \
		"setenv bootargs ${x_bootargs} ${x_bootargs_root}; " \
	"fi; " \
	"bootm 0x6400000; "

 #ifdef CONFIG_NAND_HYNIX
  #ifdef CONFIG_RCVRBTN
   #define CONFIG_EXTRA_ENV_SETTINGS	\
	"x_usbcmd_kernel=ext2load usb 0 0x6400000 /boot/uImage\0" \
	"x_usbarg_root=root=/dev/sda1 rw noinitrd rootdelay=5 rootfstype=ext3 rootwait\0" \
	"ipaddr=192.168.10.234\0"	\
	"serverip=192.168.10.150\0"	\
	"x_bootcmd_kernel=nand read.e 0x6400000 0x400000 0x400000\0" \
	"x_bootcmd_kernel_rcvr=nand read.e 0x6400000 0x800000 0x400000\0" \
	"x_bootargs=console=ttyS0,115200\0" \
	"x_bootargs_root=ubi.mtd=5 root=ubi0:rootfs rootfstype=ubifs\0" \
	"x_bootargs_root_rcvr=ubi.mtd=3 root=ubi0:rootfs rootfstype=ubifs\0"
  #else
   #define CONFIG_EXTRA_ENV_SETTINGS	\
	"x_usbcmd_kernel=ext2load usb 0 0x6400000 /boot/uImage\0" \
	"x_usbarg_root=root=/dev/sda1 rw noinitrd rootdelay=5 rootfstype=ext3 rootwait\0" \
	"ipaddr=192.168.10.234\0"	\
	"serverip=192.168.10.150\0"	\
	"x_bootcmd_kernel=nand read.e 0x6400000 0x400000 0x400000\0" \
	"x_bootargs=console=ttyS0,115200\0"	\
	"x_bootargs_root=ubi.mtd=3 root=ubi0:rootfs rootfstype=ubifs\0"
  #endif
 #else
  #define CONFIG_EXTRA_ENV_SETTINGS \
	"x_usbcmd_kernel=ext2load usb 0 0x6400000 /boot/uImage\0" \
	"x_usbarg_root=root=/dev/sda1 rw noinitrd rootdelay=5 rootfstype=ext3 rootwait\0" \
	"ipaddr=192.168.10.234\0" \
	"serverip=192.168.10.150\0" \
	"x_bootcmd_kernel=nand read.e 0x6400000 0x100000 0x400000\0" \
	"x_bootargs=console=ttyS0,115200\0" \
	"x_bootargs_root=ubi.mtd=3 root=ubi0:rootfs rootfstype=ubifs\0"
 #endif
#endif
	
/*
 * Ethernet Driver configuration
 */
#ifdef CONFIG_CMD_NET
#define CONFIG_NETCONSOLE
#undef CONFIG_PHYLIB		
#undef CONFIG_PHY_MARVELL	
#define CONFIG_PHY_GIGE		1
#undef CONFIG_SYS_FAULT_ECHO_LINK_DOWN/* FIXME: Skip link check as ports are swapped */
#define CONFIG_MVGBE_PORTS     	{1, 1}  /* enable both ports */
#define CONFIG_PHY_BASE_ADR    	0

#endif /* CONFIG_CMD_NET */

#define CONFIG_SYS_ALT_MEMTEST

#if defined(CONFIG_RCVRBTN) || defined(CONFIG_BOARD_IS_PROMWAD_THINCLIENT)
#define BOARD_LATE_INIT
#endif

/*
 * RTC driver configuration
 */
#ifdef CONFIG_CMD_DATE
#define CONFIG_RTC_MV
#endif /* CONFIG_CMD_DATE */

#endif /* _CONFIG_WTPLUG_H */
